%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File name: GenerateDataDelay.m
%
% Description: Generates data using a general nonlinear 
%              state space model. The model is assumed to be 
%              time invariant with delay added by interpolation 
%              afterwards
%
% Inputs: Input signals, dimensions of input, system and
%         outputs. Also the input type may be selected - in
%         this case the input is generated automatically.
%         Delay selection is added in this latest release.
%
% Outputs: Simulated output signals
%
% Author and copyright: Torbjrn Wigren
%                       Systems and Control
%                       Department of Information Technology
%                       Uppsala University
%                       Uppsala
%                       Sweden
%
% Revision log: Created 2003-04-27.
%               Delay simulation added 2017-01-25.
%
% Firm revision: revision 1.0, 2004-12-23.
%                revision 2.0, 2005-08-24 (no changes).
%                revision 3.0, 2017-01-25. 
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%
% Compute input signals
%

SamplingInstances=(Ts:Ts:N*Ts);
CompleteClockCycles=floor(N./ClockPeriod);
RemainderSamples=N-CompleteClockCycles.*ClockPeriod;
u=zeros(nu,N);

for iu=1:nu % Loop over inputs
   switch InputType(iu,:)
   case 'PRBS       '
      temp=uAmplitude(iu,1)*sign(randn(1,CompleteClockCycles(iu,1)+1))+uMean(iu,1); % Basic input signal component
      u(iu,1:ClockPeriod(iu,1)*CompleteClockCycles(iu,1))=kron(temp(1,1:CompleteClockCycles(iu,1)),ones(1,ClockPeriod(iu,1))); % Spread
      if RemainderSamples(iu,1)>0
         u(iu,ClockPeriod(iu,1)*CompleteClockCycles(iu,1)+1:N)=kron(temp(1,CompleteClockCycles(iu,1)+1),ones(1,RemainderSamples(iu,1))); % Add final samples
      end
   case 'Gaussian   '
      temp=uAmplitude(iu,1)*randn(1,CompleteClockCycles(iu,1)+1)+uMean(iu,1); % Basic input signal component
      u(iu,1:ClockPeriod(iu,1)*CompleteClockCycles(iu,1))=kron(temp(1,1:CompleteClockCycles(iu,1)),ones(1,ClockPeriod(iu,1))); % Spread
      if RemainderSamples(iu,1)>0
         u(iu,ClockPeriod(iu,1)*CompleteClockCycles(iu,1)+1:N)=kron(temp(1,CompleteClockCycles(iu,1)+1),ones(1,RemainderSamples(iu,1))); % Add final samples
      end
   case 'UniformPRBS'
      temp1=sign(randn(1,CompleteClockCycles(iu,1)+1)); % Basic input signal component
      temp2=uAmplitude(iu,1)*(2*rand(1,CompleteClockCycles(iu,1)+1)-1); % Compute amplitude distribution values      
      temp=temp1.*temp2+uMean(iu,1);
      u(iu,1:ClockPeriod(iu,1)*CompleteClockCycles(iu,1))=kron(temp(1,1:CompleteClockCycles(iu,1)),ones(1,ClockPeriod(iu,1))); % Spread
      if RemainderSamples(iu,1)>0
         u(iu,ClockPeriod(iu,1)*CompleteClockCycles(iu,1)+1:N)=kron(temp(1,CompleteClockCycles(iu,1)+1),ones(1,RemainderSamples(iu,1))); % Add final samples
      end
   case 'SineWave   '
      u(iu,1:N)=uAmplitude(iu,1)*sin((Ts:Ts:N*Ts)*uFrequency(iu,1))+uMean(iu,1);      
   case 'Custom     ' % just read
      u=InputSignal;
   end
end   

%
% Compute system disturbances
%

w=zeros(nx_0,N);
for ix=1:nx_0
   switch DisturbanceTypeSystem(ix,:)
   case 'WGN     '
      w(ix,1:N)=wSigma(ix,1)*randn(1,N)+wMean(ix,1);
   case 'SineWave'
      w(ix,1:N)=wSineAmplitude(ix,1)*sin((Ts:Ts:N*Ts)*wSineFrequency(ix,1))+wMean(ix,1);
   case 'Custom  ' % just read
      w=SystemDisturbance;
   end
end

%
% Compute measurement disturbances
%

e=zeros(ny,N);
for iy=1:ny
   switch DisturbanceTypeMeasurement
   case 'WGN     '
      e(iy,1:N)=eSigma(iy,1)*randn(1,N)+eMean(iy,1);
   case 'SineWave'
      e(iy,1:N)=eSineAmplitude(iy,1)*sin((Ts:Ts:N*Ts)*eSineFrequency(iy,1))+eMean(iy,1);
   case 'Custom  ' % just read
      e=MeasurementDisturbance;
   end
end

%
% Integrate 
%

% Apply an Euler forward method...

x=zeros(nx_0,N);
y=zeros(ny,N);
yT=zeros(ny,N);
x(:,1)=x0;
t=SamplingInstances(1,1);
percentReady=0;
for t=SamplingInstances(1,1):Ts:SamplingInstances(1,N-1)
   if round(t*100/N/Ts)>percentReady
      percentReady=round(t*100/N/Ts)
   end
   k=round(t/Ts); % Sampling index into matrices
   x(:,k+1)=x(:,k)+Ts*f(t,x(:,k),u(:,k),w(:,k));
   y(:,k)=h(t,x(:,k),u(:,k),e(:,k));
end

% Generate delayed output signal

d=floor((Td+1e-12)/Ts);
if Td>0
   for t=SamplingInstances(1,1):Ts:SamplingInstances(1,N-1)
      k=round(t/Ts); % Sampling index into matrices
      if k<=d+1
         yT(:,k)=zeros(size(y,1),1);
      else
         yT(:,k)=(1-(Td/Ts-d))*h(t,x(:,k-d),u(:,k-d),e(:,k-d))+(Td/Ts-d)*h(t,x(:,k-d-1),u(:,k-d+1),e(:,k-d+1));
      end
   end    
else
    yT=y;
end

% Compute derivative of input signal - for use in gradient computations

dudt=diff(u,1,2);
dudt=[dudt dudt(:,size(dudt,2))]/Ts;
