%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% File name: RPEMDelay.m
%
% Description: MIMO RPEM algorithm for general nonlinear 
%              identification. Include noise covariance
%              estimation. Delay estimation added.
%
% Inputs: Input signals, dimensions and output signals. 
%
% Outputs: Estimated parameters.
%
% Author and copyright: Torbjrn Wigren
%                       Systems and Control
%                       Department of Information Technology
%                       Uppsala University
%                       Uppsala
%                       Sweden
%
% Revision log: Created 2003-04-27.
%				Fixed output equation error, time index of u 
%               changed from t to t+1, 2005-08-23.
%               Added delay estimation in revision 3.0 2017-01-25.
%
% Firm revision: revision 1.0, 2004-12-23.
%                revision 2.0, 2005-08-23.
%                revision 3.0, 2017-01-27.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%
% Dimensions
% 

%  PolynomialOrders=zeros(nx+nu,1);
%
%  Example:
%
%  polynomialOrders=[3 1 3]';
%  notUsedIndices=[
%   1 1 1;
%   2 1 1;
%   3 1 1];
%
%  Note: First index corresponds to x_1 (slowest variation), the 
%        following to x_2 and so on.
%

%
% Generate indices to sum over
%

if ~exist('theta_0_new')
   'Input polynomialOrders and notUsedIndices'
   keyboard;
   allIndices=GenerateIndices(polynomialOrders,notUsedIndices)
   LinearizedPoleRadii=eye(nx);
   while any(LinearizedPoleRadii>stabilityLimit) % Neglect delay for initiaization
	  'Input theta_0'
 	  keyboard;
  	  SM=eye(nx)+scalingTs*Ts*dfdx_m(x_m_0,u(:,1),theta_0,allIndices);
      LinearizedPoleRadii=abs(eig(SM))
   end  
end
if exist('theta_0_new')
   allIndices=GenerateIndices(polynomialOrders,notUsedIndices);
   theta_0=theta_0_new;
end

%
% Input delay
%

if ~exist('Td_0_new')
   'Input Td_0'
   keyboard 
   while ~((Td_0>=Td_min) & (Td_0<= Td_max))
      'Input Td_0'
      keyboard
   end
end
if exist('Td_0_new')
   Td_0=Td_0_new;
end

%
% Get the right output signal
%

y=yT;

%
% Set up parameter vector
%

theta_0=[Td_0; theta_0]; 
dimTheta=size(allIndices,1)+1;

%
% Remaining initial values
%

Gamma_0=initialNoiseVariance*eye(size(y,1),size(y,1));
R_0=scaleFactorR*eye(dimTheta,dimTheta);
R_0(1,1)=scaleFactorRDelay;
psi_l_0=zeros(dimTheta-1,nx);
psi_Td_0=zeros(nx,1);
psi_x_Td_0=zeros(dimTheta,1);

% Storage variables

x_m_save=zeros(nx,ceil(N/downSampling)+1);
y_m_save=zeros(ny,ceil(N/downSampling)+1);
perr_save=zeros(ny,ceil(N/downSampling));
theta_save=zeros(dimTheta,ceil(N/downSampling)+1);
eig_R_save=zeros(dimTheta,ceil(N/downSampling)+1);
condition_save=zeros(1,ceil(N/downSampling)+1);

%
% Initiation of variables
% 

% Variables for delay handling

u_m_d=[];
dudt_m_d=[];
y_m_d=[];
x_m_d=[];
psi_l_d=[];
psi_Td_d=[];
H_d=[];
G_d=[];
for i=1:nd
    u_m_d=[u_m_d u_m_0];
    dudt_m_d=[dudt_m_d dudt_m_0];
    y_m_d=[y_m_d y_m_0];
    x_m_d=[x_m_d x_m_0];
    psi_l_d=[psi_l_d psi_l_0];
    psi_Td_d=[psi_Td_d psi_Td_0];
    H_d=[H_d dhdx_m(x_m_0,u_m_0)];
    G_d=[G_d dhdu_m(x_m_0,u_m_0)];
end;

% Algorithmic variables

mu=mu_0;
u_m=u_m_0;
dudt_m=dudt_m_0;
Gamma=Gamma_0;
R=R_0;
theta=theta_0;
x_m=x_m_0;
y_m=y_m_0;
psi_l=psi_l_0;
psi_Td=psi_Td_0;
psi_x_Td=psi_x_Td_0;

% Storage variables

x_m_save(:,1)=x_m_0;
y_m_save(:,1)=y_m_0;
theta_save(:,1)=theta_0;

% Algorithmic loop

percentReady=0;

for t=1:N % Main loop starts...
   
   if round(t*100/N)>percentReady
      percentReady=round(t*100/N)
   end
   
   %
   % Calculate mu(t)
   %
   
   mu=mu0*mu+1-mu0;
   
   %
   % Calculate delayed u and dudt
   %
   
   for i=1:nd
       if (t-i>=0)
          u_m_d(:,i)=u(:,t-i+1);
          dudt_m_d(:,i)=dudt(:,t-i+1);
       else
          u_m_d(:,i)=u(:,1);
          dudt_m_d(:,i)=dudt(:,1);
       end
   end
      
   %
   % Prediction error update
   %
   
   perr=y(:,t)-y_m;
   
   %
   % Perform tentative updates
   %
   
   GammaProjected=Gamma;
   RProjected=R;
   thetaProjected=theta;
   
   Gamma=Gamma+(mu/(muFactor+t))*(perr*perr'-Gamma);
   R=R+(mu/(muFactor+t))*(psi_x_Td*inv(Gamma)*psi_x_Td'-R);
   theta=theta+(mu/(muFactor+t))*inv(R)*psi_x_Td*inv(Gamma)*perr;

   %
   % Model and gradient stability check
   %
   
   SM=eye(nx)+scalingTs*Ts*dfdx_m(x_m,u_m,theta(2:dimTheta,1),allIndices);
   LinearizedPoles=eig(SM);
   if all(abs(LinearizedPoles)<stabilityLimit)
      stable=1;
   else 
      stable=0;
   end
   
   %
   % Check if the delay is within range
   %
   
   if ((theta(1,1)>=Td_min) & (theta(1,1) < Td_max))
       delayInRange=1;
   else
       delayInRange=0;
   end

   %
   % Perform projection - if needed
   %
  
   if (stable & ~delayInRange)
       theta(1,1)=thetaProjected(1,1);
   end
   if (~stable & delayInRange)
       Gamma=GammaProjected;
       R=RProjected;
       theta(2:dimTheta,1)=thetaProjected(2:dimTheta,1);
   end
   if (~stable & ~delayInRange)
       Gamma=GammaProjected;
       R=RProjected; 
       theta=thetaProjected;
   end
 
   %
   % Gradient computation
   %
   
   % Interpolation weights
   
   Td = theta(1,1);
   intDelay=floor(Td/Ts);
   alpha=(Td-intDelay*Ts)/Ts;

   % Delayed recursive dynamic gradient components
   
   for i=nd:-1:2
       psi_l_d(:,(i-1)*nx+1:i*nx)=psi_l_d(:,(i-2)*nx+1:(i-1)*nx);
   end
   psi_l_d(:,1:nx)=psi_l_d(:,1:nx)+scalingTs*Ts*(psi_l_d(:,1:nx)*dfdx_m(x_m_d(:,1),u_m_d(:,1),theta(2:dimTheta,1),allIndices)'+dfdtheta_m(x_m_d(:,1),u_m_d(:,1),theta(2:dimTheta),allIndices)');
   
   psi_l=(1-alpha)*psi_l_d(:,intDelay*nx+1:(intDelay+1)*nx)+alpha*psi_l_d(:,(intDelay+1)*nx+1:(intDelay+2)*nx);
   
   % Delayed delay gradient 
   
   for i=nd:-1:2
      psi_Td_d(:,i)=psi_Td_d(:,i-1);
   end
   psi_Td_d(:,1)=f_m(x_m_d(:,1),u_m_d(:,1),theta(2:dimTheta,1),allIndices);

   psi_Td=(1-alpha)*psi_Td_d(:,intDelay+1)+alpha*psi_Td_d(:,intDelay+2);
   
   % Delayed H and G 
   
   for i=nd:-1:2
      H_d(:,(i-1)*nx+1:i*nx)=H_d(:,(i-2)*nx+1:(i-1)*nx);
      G_d(:,(i-1)*nu+1:i*nu)=G_d(:,(i-2)*nu+1:(i-1)*nu);
   end
   H_d(:,1:nx)=dhdx_m(x_m_d(:,1),u_m_d(:,1));
   G_d(:,1:nu)=dhdu_m(x_m_d(:,1),u_m_d(:,1));
  
   H=(1-alpha)*H_d(:,intDelay*nx+1:(intDelay+1)*nx)+alpha*H_d((intDelay+1)*nx+1:(intDelay+2)*nx);
   G=(1-alpha)*G_d(:,intDelay*nu+1:(intDelay+1)*nu)+alpha*G_d((intDelay+1)*nu+1:(intDelay+2)*nu);
   
   % Delayed u and dudt 
   
   u_m=(1-alpha)*u_m_d(:,intDelay+1)+alpha*u_m_d(:,intDelay+2);
   dudt_m=(1-alpha)*dudt_m_d(:,intDelay+1)+alpha*dudt_m_d(:,intDelay+2);
   
   % Final gradient 
   
   psi_x_Td = [-H*psi_Td-G*dudt_m;psi_l*H'];

   %
   % State update
   %
   
   for i=nd:-1:2
      x_m_d(:,i)=x_m_d(:,i-1);
   end
   x_m_d(:,1)=x_m_d(:,1)+scalingTs*Ts*f_m(x_m_d(:,1),u_m_d(:,1),theta(2:dimTheta,1),allIndices);

   x_m=(1-alpha)*x_m_d(:,intDelay+1)+alpha*x_m_d(:,intDelay+2);
   

   %   
   % Delayed u(:,t+1) 
   %
   
   if t<N
      if (intDelay>0)
         uNext_m=(1-alpha)*u_m_d(:,intDelay)+alpha*u_m_d(:,intDelay+1);
      else
         uNext_m=(1-alpha)*u(:,t+1)+alpha*u_m_d(:,intDelay+1);
      end
      
   %
   % Model output
   %
      
      y_m=h_m(x_m,uNext_m);       
   end
     
   %
   % Check and save
   %
   
   if rem(t,downSampling)==0 % Time to save
      x_m_save(:,t/downSampling+1)=x_m;
      y_m_save(:,t/downSampling+1)=y_m;
      perr_save(:,t/downSampling)=perr;
      theta_save(:,t/downSampling+1)=theta;
      eig_R_save(:,t/downSampling+1)=eig(R);
      condition_save(:,t/downSampling+1)=cond(R);
   end
   
end

reInitiateTheta=theta;
reInitiateIndices=allIndices;

%
% Compute the re-scaled, original parameter vector
%

rescaledTheta=theta;
for i=2:dimTheta
   for j=1:nx
      thisSubIndex=allIndices(i-1,j);
      rescaledTheta(i,1)=rescaledTheta(i,1)*((1/scalingTs)^(thisSubIndex*(j-1)));
   end
end
rescaledTheta(2:dimTheta,1)=rescaledTheta(2:dimTheta)*(scalingTs)^nx;

%
% End of algorithmic loop
%

[theta]

